import sys
import threading
from queue import Queue, Empty 
import time
import code

import arcade
import pie4t
from . import common

import __main__
from __main__ import __dict__ as main_dict

class Repl:
    """  Repl main and readline thread """
    def start_repl(self): 
        self.cmd_queue = Queue()
        self.is_cmd_completed = True

        self.console = code.InteractiveConsole(locals=main_dict,filename='輸入')   
        
        arcade.schedule(self.handle_repl, 0.2)
        t = threading.Thread(target=self.readline_thread)
        common.repl_thread = t
        t.daemon = True
        t.start()  

    def readline_thread(self):
        print('物理模擬開始 ')
        print('\n')
        print('>>> ', end='')
        while True:
             try:
                line = sys.stdin.readline()
                self.cmd_queue.put(line)
                time.sleep(0.2)
             except RuntimeError as e :
                 print('請按上方紅STOP按鈕重啟互動環境')
                 return

    def handle_repl(self, dt):
        try:
            line = self.cmd_queue.get(block=False)
        except Empty:
            return

        # strip enter  or  leading white space
        line = line.rstrip()
        if self.is_cmd_completed: line = line.lstrip()


        if self.console.push(line):
            # not complele
            print('... ', end='')
            self.is_cmd_completed = False
        else:
            # complete
            print('>>> ', end='')
            self.is_cmd_completed = True

        # try:
        #     if '=' in line:
        #         #exec(line)
        #         exec(line, main_dict)
        #     else:
        #         #r = eval(line)
        #         r = eval(line, main_dict)
        #         if r:
        #             print(r)
        #     #print('Got: ', line[:-1], '---')
        
        # except Exception as e:
        #     print(e)
        # finally:
        #     print('4t>>> ', end='') 


from time import time

import arcade
import pymunk
from pymunk.vec2d import Vec2d
from . import common

class DotMark:
    def __init__(self, x=0, y=0):
        self.enabled = False

        self.x = x
        self.y = y

        self.timestamp = time()
        

    def lazy_setup(self):
        self.shape_list = arcade.ShapeElementList()
        
        s = arcade.create_ellipse(0, 0 , 20, 20, (255,255,255,120))
        self.shape_list.append(s)

        vs = arcade.create_rectangle(0,0, 3, 20, arcade.color.BLACK )
        hs = arcade.create_rectangle(0,0, 20, 3, arcade.color.BLACK )
        self.shape_list.append(vs)
        self.shape_list.append(hs)

    def update_pos(self, x, y):
        self.x = x
        self.y = y
        self.timestamp = time()
        self.enabled = True
        

    def draw(self):
        if self.enabled:
            if time() - self.timestamp < common.ASSIST_MARK_PERIOD:
                self.shape_list.center_x = self.x
                self.shape_list.center_y = self.y
                self.shape_list.angle = 0 # no rotation
                self.shape_list.draw()
            else: # expired
                self.enabled = False


class SegmentAddAssist:
    def __init__(self):
        self._enabled = False
        self.first_clicked = False
        self.first_x = 0
        self.first_y = 0
        self.second_x = 0 
        self.second_y = 0
        self.dirty = False
         
    def enable(self):
        self._enabled = True
        self.first_clicked = False
        self.first_x = 0
        self.first_y = 0
        self.second_x = 0 
        self.second_y = 0
        self.dirty = False
        cur = common.stage.get_system_mouse_cursor('crosshair')
        common.stage.set_mouse_cursor(cur)
        common.stage.模擬暫停 = True        

    def disable(self):
        self._enabled = False
        common.stage.set_mouse_cursor(None)
        common.stage.模擬暫停 = False

        if self.dirty:
           common.stage.save_terrain()  

    @property
    def enabled(self):
        return self._enabled

    # def lazy_setup(self):
    #     #self.shape_list = arcade.ShapeElementList()
    #     pass

    def click(self, x, y):
        if not self.first_clicked:
            # first click
            self.first_x = x
            self.first_y = y
            self.first_clicked = True
        else: # second click
            self.second_x = x
            self.second_y = y
            self.first_clicked = False

            if not (self.first_x == self.second_x and self.first_y == self.second_y):
                common.stage.新增線段((self.first_x,self.first_y),
                                      (self.second_x, self.second_y),
                                       common.SEG_THICKNESS)
                self.dirty = True
                
    def draw(self):
        if self._enabled and self.first_clicked:
            fx = self.first_x
            fy = self.first_y
            mx = common.stage.mouse_x
            my = common.stage.mouse_y
            arcade.draw_line(fx, fy, mx, my,arcade.color.GREEN ,common.SEG_THICKNESS)


class SegmentRemoveAssist:
    def __init__(self):
        self._enabled = False
        self.dirty = False
        self.hover_segment = None
        self.seg_filter = pymunk.ShapeFilter(mask=common.CATE_SEGMENT)

    def enable(self):
        self._enabled = True
        self.dirty = False
        self.hover_segment = None
        cur = common.stage.get_system_mouse_cursor('help')
        common.stage.set_mouse_cursor(cur)
        common.stage.模擬暫停 = True        

    def disable(self):
        self._enabled = False
        common.stage.set_mouse_cursor(None)
        common.stage.模擬暫停 = False

        if self.dirty:
           common.stage.save_terrain()  

    @property
    def enabled(self):
        return self._enabled

    # def lazy_setup(self):
    #     #self.shape_list = arcade.ShapeElementList()
    #     pass

    def draw(self):
        if self._enabled:
            if self.hover_segment:
                a = self.hover_segment.a
                b = self.hover_segment.b
                thickness = self.hover_segment.thickness
                arcade.draw_line(a[0], a[1], b[0], b[1],
                     arcade.color.RED, thickness)

    def click(self, x, y):
        if self._enabled:
            if self.hover_segment:
                common.stage.移除(self.hover_segment)
                self.hover_segment = None
                self.dirty = True

    def update_hover(self,x, y):
        if self._enabled:
            query = common.stage.space.point_query_nearest((x,y), 3, self.seg_filter)
            if query:
                self.hover_segment = query.shape.obj
            else:
                self.hover_segment = None


class ArrowAssist:
    def __init__(self):
        #self.mode_turn_on = False # turn on by user
        self.enabled = False  # drawing
        #self.vector = 0
        self.start_x = 0
        self.start_y = 0
        # self.mouse_x = 0
        # self.mouse_y = 0


    def start(self, pos=None):
        #if self.mode_turn_on :


        if pos is None :
            self.start_x = common.stage.mouse_x
            self.start_y = common.stage.mouse_y
        else:
            pos = Vec2d(pos)
            self.start_x = pos.x
            self.start_y = pos.y
            

        self.enabled = True

    # def update_mouse_pos(self, x, y):
    #     if self.mode_turn_on and self.enabled:
    #         self.mouse_x = x
    #         self.mouse_y = y

    def launch(self):
        if self.enabled:
            self.enabled = False




    # def lazy_setup(self):
    #     pass

    def draw(self):
        if self.enabled :
            sx = self.start_x
            sy = self.start_y
            mx = common.stage.mouse_x
            my = common.stage.mouse_y
            length = self.vector.length
            #print("length: ",self.vector.length )
            if length > 40:
                line_v = self.vector
                line_v.length -= 30

                # line
                arcade.draw_line(sx, sy, sx + line_v.x, sy + line_v.y,
                            arcade.color.RED ,common.SEG_THICKNESS*3)
                # triangle
                #if length 
                if length < 500:
                    degree_delta =  15 - length / 50
                else:
                    degree_delta = 5
                

                left_v = self.vector
                left_v.length -= 30
                left_v.angle_degrees += degree_delta 

                right_v = self.vector
                right_v.length -= 30
                right_v.angle_degrees -= degree_delta 

                

                arcade.draw_triangle_filled(mx, my,
                    sx + left_v.x, sy + left_v.y,
                    sx + right_v.x, sy + right_v.y,
                    arcade.color.RED
                )

            
            

    @property
    def vector(self):
            sx = self.start_x
            sy = self.start_y
            mx = common.stage.mouse_x
            my = common.stage.mouse_y
            return Vec2d(mx - sx, my - sy)

    @property
    def start_pos(self):
        return Vec2d(self.start_x, self.start_y)



class CoordinateAssist:
    def __init__(self):
        self._enabled = False
        self.shape_element = None
        self.text_list = []
        self.coor_start = 0
        self.win_width = common.stage.win_width
        self.win_height = common.stage.win_height
        self.font =  ('C:/Windows/Fonts/msjh.ttc','arial')

        upper_bound = max(self.win_width, self.win_height)
        upper_bound = (upper_bound + 99) // 100 * 100 

        self.coor_end = upper_bound
        self.coor_step = 50 
        self.label_step = 100

    def enable(self):
        self._enabled = True
        

    def disable(self):
        self._enabled = False
        

    @property
    def enabled(self):
        return self._enabled

    def lazy_setup(self):
        self.shape_element = arcade.ShapeElementList()
        i = 0
        for y in range(self.coor_start, self.coor_end + self.coor_step, self.coor_step):
            l = arcade.create_line(0, y, self.win_width ,y , arcade.color.ANTIQUE_BRONZE  , 1)
            self.shape_element.append(l)

            r = arcade.create_rectangle(0, y, 8, 3, arcade.color.WHITE_SMOKE)
            self.shape_element.append(r)
            

        # Draw the x labels.
        i = 0
        for x in range(self.coor_start, self.coor_end + self.coor_step, self.coor_step):
            l = arcade.create_line(x, 0, x ,self.win_height , arcade.color.ANTIQUE_BRONZE  , 1)
            self.shape_element.append(l)

            r = arcade.create_rectangle(x, 0, 3, 8, arcade.color.WHITE_SMOKE)
            self.shape_element.append(r)

        # ---------------------------

        for y in range(self.coor_start, self.coor_end + self.label_step, self.label_step):            
            t = arcade.Text(f"{y}", 5, y+3, arcade.color.WHITE_SMOKE, 12, anchor_x="left", 
                            anchor_y="center", font_name=self.font)
            self.text_list.append(t)
            i += 1

        i = 1
        for x in range(self.coor_start + self.label_step, self.coor_end + self.label_step, self.label_step):    
            t = arcade.Text( f"{x}" , x, 5, arcade.color.WHITE_SMOKE, 12, anchor_x="center", 
                            anchor_y="bottom", font_name=self.font)
            self.text_list.append(t)
            i += 1 

        t = arcade.Text("Y", 25, common.stage.win_height-25, arcade.color.WHITE_SMOKE, 18, 
                        anchor_x="center", anchor_y="center", font_name=self.font)
        self.text_list.append(t)

        t = arcade.Text("X", common.stage.win_width-25, 25, arcade.color.WHITE_SMOKE, 18, 
                        anchor_x="center", anchor_y="center", font_name=self.font)
        self.text_list.append(t)



    def draw(self):
        if self._enabled:
            self.shape_element.center_x = 0
            self.shape_element.center_y = 0
            self.shape_element.angle = 0
            self.shape_element.draw()

            for t in self.text_list:
                t.draw()

            # i = 0
            # for y in range(self.coor_start, self.coor_end + self.label_step, self.label_step):            
            #     arcade.draw_text(f"{y}", 5, y+3, arcade.color.WHITE_SMOKE, 12, anchor_x="left", 
            #                     anchor_y="center", font_name=self.font)
            #     i += 1

            # i = 1
            # for x in range(self.coor_start + self.label_step, self.coor_end + self.label_step, self.label_step):    
            #     arcade.draw_text( f"{x}" , x, 5, arcade.color.WHITE_SMOKE, 12, anchor_x="center", 
            #                     anchor_y="bottom", font_name=self.font)
            #     i += 1 

            # arcade.draw_text("Y", 25, common.stage.win_height-25, arcade.color.WHITE_SMOKE, 18, 
            #                 anchor_x="center", anchor_y="center", font_name=self.font)
            # arcade.draw_text("X", common.stage.win_width-25, 25, arcade.color.WHITE_SMOKE, 18, 
            #                 anchor_x="center", anchor_y="center", font_name=self.font)

  

            





from . import common
from .engine import PhysicsEngine

物理引擎 = PhysicsEngine

__all__ = [ 
            'add_circle', '新增圓球', '物理引擎',
            'simulate', '模擬主迴圈','模擬進行中', '移除',
            'add_segment','新增線段','add_box','新增方塊',



            ]



######## top level function

def add_circle(*args, **kwargs):
    if not common.is_engine_created:
        PhysicsEngine()
    return common.stage.新增圓球(*args, **kwargs)

新增圓球 = add_circle


def add_box(*args, **kwargs):
    if not common.is_engine_created:
        PhysicsEngine()
    return common.stage.新增方塊(*args, **kwargs)

新增方塊 = add_box

def add_segment(*args, **kwargs):
    if not common.is_engine_created:
        PhysicsEngine()
    return common.stage.新增線段(*args, **kwargs)

新增線段 = add_segment



def 移除(obj):
    if not common.is_engine_created:
        PhysicsEngine()
    return common.stage.移除(obj)



def simulate():
    if not common.is_engine_created:
        PhysicsEngine()

    common.stage.simulate()


模擬主迴圈 = simulate

模擬進行中 = simulate

#def module_init():
    # module reference
    #__main__.pie4t_module = sys.modules['pie4t']

    ## check batch mode
    # had __file__ and import pie4t in files
    # if hasattr(__main__ , '__file__') and __main__.__file__.endswith('py'):
    #     #print('has __file__')
    #     try:
    #         with open(__main__.__file__, encoding='utf8') as f:
    #             lines = f.readlines()
    #             for i in lines:
    #                 if 'import' in i and 'pie4t' in i:
    #                     __main__.pie4t_module.is_batch_mode = True
                        
    #                     print(__file__, ': is batch mode,')
    #                     break
    #     except FileNotFoundError:
    #         pass

    # create game win
    #__main__.pie4t_module.game_win = PhysicsEngine()
    #__main__.pie4t_module.is_initialized = True

### module init

#module_init()



if __name__ == '__main__' :
    pass
    



import random
import math

import pymunk
from pymunk.vec2d import Vec2d
import arcade

from . import common
from .common import SHAPE_COLORS
from .physics_common import PhysicsCommon

class Box(PhysicsCommon):
    def __init__(self, 寬=None, 高=None):
        # pymunk part
        #self.density = 1
        self.lazy_setup_done = False
        self.type_ = '方塊'

        self.phy_body = pymunk.Body(body_type=pymunk.Body.DYNAMIC)
        self.phy_body.velocity_func = self.limit_velocity
        self.phy_body.position_func = self.limit_position
      

        if 寬 is not None :
            self.box_width = round(max(寬, 4),0)
        else:
            self.box_width = random.randrange(common.BOX_WIDTH_MIN,
                                common.BOX_WIDTH_MAX)

        if 高 is not None :
            self.box_height = round(max(高, 4),0)
        else:
            self.box_height = random.randrange(common.BOX_WIDTH_MIN,
                                common.BOX_WIDTH_MAX)


        self.phy_shape = pymunk.Poly.create_box(self.phy_body, 
                                (self.box_width, self.box_height), radius=0)


        win_width = common.stage.win_width
        win_height = common.stage.win_height
        x = random.randint(int(win_width*0.25), int(win_width*0.75))
        y = random.randint(int(win_height*0.7), int(win_height*0.9))
        self.phy_body.position = x , y

        self.phy_shape.density = common.DENSITY
        self.phy_shape.friction = common.BOX_FRICTION
        self.phy_shape.elasticity = common.BOX_ELASTICITY
        # keep referene to self in shape. For delete
        self.phy_shape.obj = self
        self.phy_shape.filter = pymunk.ShapeFilter(categories=common.CATE_BOX)

        common.stage.space.add(self.phy_body, self.phy_shape)
        
        if common.stage.is_engine_running:
            self.lazy_setup()
            self.lazy_setup_done = True


    def lazy_setup(self):
        ### arcade part
        if not self.lazy_setup_done:
            #print('do circle lazy setup')
            self.box_color = random.choice(SHAPE_COLORS)

            #  two shape element for different body type           
            self.dynamic_shape_element = arcade.ShapeElementList()
            self.kinematic_shape_element = arcade.ShapeElementList()  

            self.make_dynamic_shape()
            self.make_kinematic_shape()

            if self.物理反應 :
                self.current_shape_element = self.dynamic_shape_element
            else :
                self.current_shape_element = self.kinematic_shape_element
            
            
    def make_dynamic_shape(self):
            # dynamic  circle
            
            s = arcade.create_rectangle_filled(0, 0, self.box_width, self.box_height,
                                        self.box_color)
            self.dynamic_shape_element.append(s)

            r = self.box_color[0]+50
            r = r if r <=255 else 255
            g = self.box_color[1]+50
            g = g if g <=255 else 255
            b = self.box_color[2]+50
            b = b if b <=255 else 255

            border_color = (r, g, b)

            s = arcade.create_rectangle(0, 0, self.box_width, self.box_height,
                                        border_color, 3, filled=False)
            self.dynamic_shape_element.append(s)
            
            s = arcade.create_rectangle_filled(int(self.box_width*0.35), 0, 
                    3, int(self.box_height*0.8),
                                        border_color)
            self.dynamic_shape_element.append(s)
            
            

    def make_kinematic_shape(self):
            # dynamic  circle

            
            s = arcade.create_rectangle_filled(0, 0, self.box_width, self.box_height,
                                        arcade.color.GRAY)
            self.kinematic_shape_element.append(s)

            s = arcade.create_rectangle(0, 0, self.box_width, self.box_height,
                                        arcade.color.LAVENDER_GRAY, 3, filled=False)
            self.kinematic_shape_element.append(s)
            
            s = arcade.create_rectangle_filled(int(self.box_width*0.35), 0, 
                    3, int(self.box_height*0.8),
                                        arcade.color.LAVENDER_GRAY)
            self.kinematic_shape_element.append(s)

            
    def __repr__(self):
        # l = round(self.box_length, 1)
        # w = round(self.box_width, 1)
        # d = round(self.phy_shape.density, 1)
        # m = round(self.phy_shape.mass, 1)
        # f = self.phy_shape.friction
        # e = self.phy_shape.elasticity

        # return f'方塊 <長{l},寬{w},密度{d},質量{m},摩擦{f},彈性{e}>'

        t = self.type_
        w = self.box_width
        h = self.box_height
        a = self.phy_shape.area
        d = self.phy_shape.density
        m = self.phy_shape.mass
        f = self.phy_shape.friction
        e = self.phy_shape.elasticity
        x = self.phy_body.position.x
        y = self.phy_body.position.y

        return f'[{t}]寬{w:.0f}, 高{h:.0f}, 面積{a:.0f}, 密度{d:.2f}, 質量{m:.1f},\n' \
               f'     摩擦{f:.1f}, 彈性{e:.1f}, x座標{x:.0f}, y座標{y:.0f}\n'



    def draw(self):
        self.current_shape_element.center_x = self.phy_body.position.x
        self.current_shape_element.center_y = self.phy_body.position.y
        self.current_shape_element.angle = math.degrees(self.phy_body.angle)
        self.current_shape_element.draw()







from inspect import signature
import sys, time, json, os

import arcade
import pymunk
import pyperclip
import pyglet

from . import common
from .common import COLLITYPE_DEFAULT, COLLITYPE_HOLE 

from .repl import Repl
from .circle import Circle
from .box import Box
from .segment import Segment
from .assist import (DotMark, SegmentAddAssist, ArrowAssist,
                        SegmentRemoveAssist, CoordinateAssist,  )

import __main__

class PhysicsEngine(arcade.Window, Repl):
    __single = None
    __first_time = True

    def __new__(clz, *args, **kwargs):
        # Singleton
        if not PhysicsEngine.__single:
            PhysicsEngine.__single = object.__new__(clz)
        else:
            print("物理舞台已存在")

        return PhysicsEngine.__single
    
    def __init__(self, *args, **kwargs):
        # check module level default physics engine
        if PhysicsEngine.__first_time:
            self.do_init(*args, **kwargs)
            PhysicsEngine.__first_time = False

    def do_init(self, 寬=common.WIN_WIDTH, 
                       高=common.WIN_HEIGHT, 
                       title=common.TITLE):    
        common.stage = self
        common.舞台 = self        
        common.is_engine_created = True
        __main__.stage = self
        __main__.舞台 = self
        __main__.key = arcade.key

        self.pause_simulate = False 
        self.slow_simulate = False

        if common.WIN_MIN_WIDTH <= 寬 <= common.WIN_MAX_WIDTH:
            self.win_width = round(寬,0)
        elif 寬 < common.WIN_MIN_WIDTH :
            self.win_width = common.WIN_MIN_WIDTH
        elif 寬 > common.WIN_MAX_WIDTH :
            self.win_width = common.WIN_MAX_WIDTH

        if common.WIN_MIN_HEIGHT <= 高 <= common.WIN_MAX_HEIGHT:
            self.win_height = round(高,0)
        elif 高 < common.WIN_MIN_HEIGHT :
            self.win_height = common.WIN_MIN_HEIGHT
        elif 高 > common.WIN_MAX_HEIGHT :
            self.win_height = common.WIN_MAX_HEIGHT

        #self.win_width = round(寬,0) if 寬 > common.MIN_WIDTH else common.MIN_WIDTH
        
        
        #self.win_height = round(高,0) if 高 > common.MIN_HEIGHT else common.MIN_HEIGHT

        #print('stage size: ', self.win_width, self.win_height)
        #__main__.中央座標 = (self.win_width, self.win_height//2)

        self.title = title
        self.set_update_rate(common.DT_UPDATE)
        self.circle_list = []
        self.poly_list = []
        self.segment_list = []
        self.is_engine_running = False

        

        # status line
        self.font =  ('C:/Windows/Fonts/msjh.ttc','arial')

        if hasattr(__main__, '__file__'):
            self.terrain_filename = __main__.__file__ + '.地形'
        else:
            # in shell , do not save terrain
            self.terrain_filename = None


        # retain mouse postion while mouse motion event
        self.mouse_x = 0
        self.mouse_y = 0

        # pymunk space
        self.space = pymunk.Space()
        self.space.gravity = common.GRAVITY
        self.sleep_time_threshold = 1
        
        self.obj_filter = pymunk.ShapeFilter(
                    mask=common.CATE_CIRCLE | common.CATE_BOX)

        # info 
        # self.info = {}
        # self.info['gravity_x'] = 0
        # self.info['gravity_y'] = 0 
        # self.info['mouse_x'] = 0
        # self.info['mouse_y'] = 0
        # self.info['obj_num'] = 0
        # self.info_text = ''
        #self.info_update()

        # infomation update (can't to fast because text create burden)
        #arcade.schedule(self.info_update, 0.4)


        # assist
        self.dot_mark = DotMark()
        self.seg_add_assist = SegmentAddAssist()
        self.seg_remove_assist = SegmentRemoveAssist()
        self.arrow_assist = ArrowAssist()
        self.coor_assist = CoordinateAssist()

        hole_handler = self.space.add_collision_handler(COLLITYPE_DEFAULT,
                COLLITYPE_HOLE)



        hole_handler.begin = self.hole_begin_callback
        hole_handler.pre_solve = self.hole_pre_solve_callback
        hole_handler.post_solve = self.hole_post_solve_callback
        hole_handler.separate = self.hole_separate_callback




        # custom event handler 
        self.user_mouse_press_handler = None
        self.user_mouse_release_handler = None
        self.user_key_press_handler = None
        self.user_key_release_handler = None
        self.user_arrow_launch_handler = None
        self.user_mouse_drag_handler = None
        self.user_object_click_handler = None
        self.user_update_handler = None
        # self.user_mouse_press_handler = lambda x, y :  None
        # self.user_mouse_release_handler = lambda x, y: None
        # self.user_key_press_handler = lambda key: None
        # self.user_key_release_handler = lambda key: None
        # self.user_arrow_launch_handler = lambda vector,start_pos :  None

        print(f"建立舞台(寬{self.win_width}x高{self.win_height})")
        

        if not self.terrain_filename or not self.load_terrain():
            #default terrain
            self.新增線段((50, 50),(self.win_width-50, 50),6)
            print('使用預設地形')

    # def info_update(self, dt=0):
    #     self.info['gravity_x'] = int(self.space.gravity.x)
    #     self.info['gravity_y'] = int(self.space.gravity.y) 
    #     self.info['obj_num'] = self.object_num

    #     gx = self.info['gravity_x']
    #     gy = self.info['gravity_y']
    #     mx = self.info['mouse_x']
    #     my = self.info['mouse_y']
    #     obj_num = self.info['obj_num']
    #     self.info_text = f'重力[{gx},{gy}] 滑鼠[{mx},{my}] 物體數:{obj_num} ' 
    #     print(self.info_text)      


    def hole_begin_callback(self, arbiter, space, data):
        #print('begin res ', arbiter.restitution)

        #print('begin ', end='')
        #shape, _ = arbiter.shapes
        #self.移除(shape.obj)
        #return False
        return True

    def hole_pre_solve_callback(self, arbiter, space, data):
        #arbiter.restitution = 1
        print('pre solve sur v ', arbiter.surface_velocity)
        #print('begin friction', arbiter.is_first_contact)
        #arbiter.friction = 0
        #print('pre_solve ', end='')

        #return False
        return True

    def hole_post_solve_callback(self, arbiter, space, data):
        #print('post_solve ', end='')
        #return True
        pass

    def hole_separate_callback(self, arbiter, space, data):
        #print('separate removal ', arbiter.is_removal)
        #return True
        pass

    def lazy_setup(self):
        super().__init__(self.win_width, self.win_height, self.title)
        #print( self.get_size())
        #print(f"舞台寬{self.win_width}x高{self.win_height}")

        #print('do engine lazy setup')
        for i in self.circle_list:
            i.lazy_setup()

        for i in self.poly_list:
            i.lazy_setup()

        for i in self.segment_list:
            i.lazy_setup()

        # assist 
        self.dot_mark.lazy_setup()
        self.coor_assist.lazy_setup()
        
        # terrain text
        self.add_terrain_text = arcade.Text('新增地形(滑鼠右鍵選定起點終點)', 50, 70, 
                            arcade.csscolor.YELLOW, 14, font_name=self.font)

        self.remove_terrain_text = arcade.Text('移除地形(滑鼠右鍵)', 50, 70, 
                            arcade.csscolor.YELLOW, 14, font_name=self.font)


    # def setup_pinball_layout(self):
        
    #     self.add_segment((488,696),(473,754), 3)
    #     self.add_segment((473,754), (437,775), 3)
    #     self.add_segment((437,775), (390,790), 3)

    #     self.add_segment((426,30),(426,660),3)

    #     # 隔板
    #     for i in range(6):
    #         self.add_segment((60+i*60,120),(60+i*60,220),3)

        

        
    #     self.add_segment((11,70),(358,47),3)
        # self.add_line()

        # self.add_line()

        # self.add_line()

        # self.add_line()

        # self.add_line()


    # def setup_wall_around(self):
    #     thick = 10
    #     #self.新增線段( (thick,25), (self.win_width-thick,25), thick)
    #     self.add_segment( (thick,self.win_height-thick), (self.win_width-thick,self.win_height-thick),thick)
    #     self.add_segment( (thick,25), (thick,self.win_height-thick),thick)
    #     self.add_segment( (self.win_width-thick,25), (self.win_width-thick,self.win_height-thick),thick)

        # pass
        # line1 = StaticLine((400,200),(100,300))
        # self.line_list.append(line1)
        # line2 = StaticLine((100,50),(50,100))
        # self.line_list.append(line2)
        
        # b = Circle()
        # b.phy_body.position = (300,150)
        # joint = pymunk.PinJoint(b.phy_body, line1.phy_body, (0,0) , (0,0))
        # joint.distance = 150
        # common.stage.space.add(joint)
        # self.circle_list.append(b)
        
        #arcade.schedule(self.add_circle, 2)
        
    def simulate(self):
        #self.setup_wall_around()
        #self.setup_pinball_layout()
        self.lazy_setup()
        self.collect_user_event_handlers()
        #self.start_repl()

        

        # try cursor
        #cur = self.get_system_mouse_cursor('crosshair')
        #cur = self.get_system_mouse_cursor('help')
        # set cursor to default
        #cur = self.get_system_mouse_cursor('help')
        self.set_mouse_cursor(None)

        self.is_engine_running = True

        #load terrain file (right after is_engine_running flag)



        arcade.run()    


    def save_terrain(self):
        if self.terrain_filename:
            terrain_list = []
            for s in self.segment_list:
                tmp = [s.a, s.b, s.thickness]
                terrain_list.append(tmp)
                
            
            with open(self.terrain_filename, 'w', encoding='utf-8') as f:
                json.dump(terrain_list, f)
                #print("已儲存地形檔")
            
        #print('save terrain')


    def load_terrain(self):
        try:
            with open(self.terrain_filename, 'r', encoding='utf-8') as f:
                terrain_list = json.load(f)
        except FileNotFoundError:
            #print('找不到地形檔')
            return False
            
        else:
            self.segment_list = []
            for item in terrain_list:
                self.新增線段(item[0], item[1], item[2])
            print('地形載入完成')
            return True

    載入地形 = load_terrain

    def collect_user_event_handlers(self):
        # if hasattr(__main__, 'on_mouse_press'):
        #     # check number of parameters
        #     sig = signature(__main__.on_mouse_press)
        #     if len(sig.parameters) == 4:
        #          # parameters: x, y, button, modifiers
        #         self.user_mouse_press_handler = __main__.on_mouse_ress
        #         print( 'handler registed: on_mouse_press' )
        #     else:
        #         print('handler error: on_mouse_press needs  4 parameters')
        #         sys.exit()

        if hasattr(__main__, '按下滑鼠時'):
            # check number of parameters
            sig = signature(__main__.按下滑鼠時)
            if len(sig.parameters) == 2:
                 # parameters: x, y, button, modifiers
                self.user_mouse_press_handler = __main__.按下滑鼠時
                print( '登錄事件函式：按下滑鼠時' )
            else:
                print('事件函式錯誤: 按下滑鼠時 需要2個參數')
                sys.exit()

        if hasattr(__main__, '拖曳滑鼠時'):
            # check number of parameters
            sig = signature(__main__.拖曳滑鼠時)
            if len(sig.parameters) == 4:
                 # parameters: x, y, button, modifiers
                self.user_mouse_drag_handler = __main__.拖曳滑鼠時
                print( '登錄事件函式：拖曳滑鼠時' )
            else:
                print('事件函式錯誤: 拖曳滑鼠時 需要4個參數')
                sys.exit()



        if hasattr(__main__, '放開滑鼠時'):
            # check number of parameters
            sig = signature(__main__.放開滑鼠時)
            if len(sig.parameters) == 2:
                 # parameters: x, y, button, modifiers
                self.user_mouse_release_handler = __main__.放開滑鼠時
                print( '登錄事件函式：放開滑鼠時' )
            else:
                print('事件函式錯誤: 放開滑鼠時 需要2個參數')
                sys.exit()

        if hasattr(__main__, '點擊物體時'):
            # check number of parameters
            sig = signature(__main__.點擊物體時)
            if len(sig.parameters) == 3:
                 # parameters: object, x, y
                self.user_object_click_handler = __main__.點擊物體時
                print( '登錄事件函式：點擊物體時' )
            else:
                print('事件函式錯誤: 點擊物體時 需要3個參數')
                sys.exit()



        if hasattr(__main__, '按下鍵盤時'):
            # check number of parameters
            sig = signature(__main__.按下鍵盤時)
            if len(sig.parameters) == 3:
                 # parameters: x, y, button, modifiers
                self.user_key_press_handler = __main__.按下鍵盤時
                print( '登錄事件函式：按下鍵盤時' )
            else:
                print('事件函式錯誤: 按下鍵盤時 需要3個參數')
                sys.exit()

        if hasattr(__main__, '放開鍵盤時'):
            # check number of parameters
            sig = signature(__main__.放開鍵盤時)
            if len(sig.parameters) == 3:
                 # parameters: x, y, button, modifiers
                self.user_key_release_handler = __main__.放開鍵盤時
                print( '登錄事件函式：放開鍵盤時' )
            else:
                print('事件函式錯誤: 放開鍵盤時 需要3個參數')
                sys.exit()

        if hasattr(__main__, '箭頭發射時'):
            # check number of parameters
            sig = signature(__main__.箭頭發射時)
            if len(sig.parameters) == 2:
                 # parameters: x, y, button, modifiers
                self.user_arrow_launch_handler = __main__.箭頭發射時
                print( '登錄事件函式：箭頭發射時' )
            else:
                print('事件函式錯誤: 箭頭發射時 需要2個參數')
                sys.exit()

        if hasattr(__main__, '當更新時'):
            # check number of parameters
            sig = signature(__main__.當更新時)
            if len(sig.parameters) == 1:
                 # parameters: dt
                self.user_update_handler = __main__.當更新時
                print( '登錄事件函式：當更新時' )
            else:
                print('事件函式錯誤: 當更新時 需要1個參數')
                sys.exit()

    ### event

    def on_draw(self):
        arcade.start_render()
        
        self.coor_assist.draw()

        for b in self.circle_list:
            b.draw()
            
        for p in self.poly_list:
            p.draw()

        for li in self.segment_list:
            #print('line: ', li.shape_element.center_x, li.shape_element.center_y)
            li.draw()


        # draw assist
        self.dot_mark.draw()
        self.seg_add_assist.draw()
        self.seg_remove_assist.draw()
        self.arrow_assist.draw()
        self.coor_assist.draw()
        # draw status line
        #gx = int(self.space.gravity.x)
        #gy = int(self.space.gravity.y)

        if self.seg_add_assist.enabled:
            self.add_terrain_text.draw()            
            # arcade.draw_text('暫停--新增地形(滑鼠右鍵)', 50, 50, 
            #                 arcade.csscolor.WHITE, 14, font_name=self.font)
        elif self.seg_remove_assist.enabled:
            self.remove_terrain_text.draw()
            # arcade.draw_text('暫停--移除地形(滑鼠右鍵)', 50, 50, 
            #                 arcade.csscolor.WHITE, 14, font_name=self.font)
    
    def on_update(self, dt):
        # physics engine 
        #print('dt:', dt)
        if not self.pause_simulate:
            if not self.slow_simulate:
                for i in range(common.DT_SPLIT_NUM):
                    self.space.step(common.DT_SPLIT)
            else: # slow mode
                for i in range(common.DT_SPLIT_NUM // 4):
                    self.space.step(common.DT_SPLIT)

            if self.user_update_handler:
                self.user_update_handler(dt)

        # step_dt = 1/200.
        # x = 0
        # while x < dt:
            
        #     x += step_dt
        #     self.space.step(step_dt)


    def on_key_press(self, symbol, mod):
        if symbol == arcade.key.ESCAPE:
            self.close()
        elif symbol in (arcade.key.LCTRL, arcade.key.RCTRL):
            self.seg_add_assist.enable()
            self.coor_assist.enable()
            
            #self.seg_remove_assist.disable()
        elif symbol in (arcade.key.LALT, arcade.key.RALT):
            self.seg_remove_assist.enable()
            self.coor_assist.enable()
            
            #self.seg_add_assist.disable()
        # elif symbol == arcade.key.TAB :
        #     self.coor_assist.enable()
        #elif not self.模擬暫停 and self.user_key_press_handler :
        elif self.user_key_press_handler :
            self.user_key_press_handler(symbol, self.mouse_x, self.mouse_y)

    def on_key_release(self, symbol, mod):
        if symbol in (arcade.key.LCTRL, arcade.key.RCTRL):
            #print('ctrl released')
            self.seg_add_assist.disable()
            self.coor_assist.disable()
        elif symbol in (arcade.key.LALT, arcade.key.RALT):
            self.seg_remove_assist.disable()
            self.coor_assist.disable()
        # elif symbol == arcade.key.TAB :
        #     self.coor_assist.disable()
        #elif not self.模擬暫停 and self.user_key_release_handler:
        elif  self.user_key_release_handler:
            self.user_key_release_handler(symbol, self.mouse_x, self.mouse_y)

    def on_mouse_press(self, x, y, button, modifiers):
        if button == arcade.MOUSE_BUTTON_RIGHT:
            if modifiers & arcade.key.MOD_CTRL:
                if self.seg_add_assist.enabled:
                    self.seg_add_assist.click(x, y)
                
                    #print('click first')
                # else: # do second point
                #     self.seg_add_assist.click_second(x, y)
                #     fx = self.seg_add_assist.first_point_x
                #     fy = self.seg_add_assist.first_point_y
                #     seg_text = "A點 = [{},{}]\nB點 = [{},{}]\n新增線段(A點, B點, 寬=4)\n"
                #     seg_text = seg_text.format(fx,fy, x, y)
                #     pyperclip.copy(seg_text)
                #     print('複製線段程式')
            elif modifiers & arcade.key.MOD_ALT:
                if self.seg_remove_assist.enabled:
                    self.seg_remove_assist.click(x, y)
            else : # no ctrl pressed
                cor_text = f'[{x},{y}]'
                pyperclip.copy(cor_text)
                print('複製座標 '+ cor_text)
                self.dot_mark.update_pos(x, y)   

        #elif button == arcade.MOUSE_BUTTON_LEFT and not self.模擬暫停:
        elif button == arcade.MOUSE_BUTTON_LEFT :
            object_clicked = False
            if self.user_object_click_handler:
                # object click has high priority
                query = common.stage.space.point_query_nearest((x,y), 0, self.obj_filter)
                if query:
                    self.user_object_click_handler(query.shape.obj, x, y)
                    object_clicked = True
            
                        
            if not object_clicked and self.user_mouse_press_handler:
            # call user define handlers
                self.user_mouse_press_handler(x, y)




    def on_mouse_drag(self, x, y, dx, dy, button, modifiers):
        if button == arcade.MOUSE_BUTTON_LEFT:
            #if not self.模擬暫停 and self.user_mouse_drag_handler:
            if self.user_mouse_drag_handler:
                self.user_mouse_drag_handler(x, y, dx, dy)


    def on_mouse_motion(self, x, y, dx, dy):
        
        # segment assist
        self.mouse_x = x
        self.mouse_y = y

        if self.seg_remove_assist.enabled:
            self.seg_remove_assist.update_hover(x, y)

        # if self.seg_add_assist.enabled:
        #     self.seg_add_assist.update_mouse_pos(x, y)

    def on_mouse_release(self, x, y, button, modifiers):
        if button == arcade.MOUSE_BUTTON_LEFT:
            # check if arrow assist need launch
            if self.arrow_assist.enabled:
                self.arrow_assist.launch()
                vector = self.arrow_assist.vector
                start_pos = self.arrow_assist.start_pos
                print('start_pos: ', start_pos)
                self.user_arrow_launch_handler(vector, start_pos)

            # call user define handlers
            #if not self.模擬暫停 and self.user_mouse_release_handler:
            if self.user_mouse_release_handler:
                self.user_mouse_release_handler(x, y)
         
    # def on_close(self):
    #     print('on_close')
    #     self.close()
    #     exit()

                

    ### add object

    def add_segment(self,*args, **kwargs):
        s = Segment(*args, **kwargs)
        self.segment_list.append(s)
        return s

    新增線段 = add_segment


    def 新增圓球(self, *args, **kwargs):
        c = Circle(*args, **kwargs)
        self.circle_list.append(c)
        return c

    def 新增方塊(self, *args, **kwargs):
        b = Box(*args, **kwargs)
        self.poly_list.append(b)
        return b

    def 移除(self, obj):

        try:

            if isinstance(obj, Segment):
                self.space.remove(obj.phy_shape)
                self.segment_list.remove(obj)
                del obj.shape_element
                del obj.phy_shape
                del obj.phy_body
                del obj
            else: 
            # remove shape and body from space
                self.space.remove(obj.phy_shape)
                self.space.remove(obj.phy_body)
                if isinstance(obj, Circle):
                    self.circle_list.remove(obj)
                    del obj.dynamic_shape_element
                    del obj.kinematic_shape_element
                
                if isinstance(obj, Box):
                    self.poly_list.remove(obj)
                    del obj.dynamic_shape_element
                    del obj.kinematic_shape_element

                del obj.phy_shape
                del obj.phy_body
                del obj
        except AttributeError:
            print('移除物體不存在')        

    ### assist
    def 箭頭開始(self, *args, **kwargs):
        self.arrow_assist.start(*args, **kwargs)


    ### property
    @property
    def gravity(self):
        return self.space.gravity

    @gravity.setter
    def gravity(self, value):
        self.space.gravity = value 

    @property
    def 重力(self):
        return self.space.gravity

    @重力.setter
    def 重力(self, value):
        self.space.gravity = value 

    @property
    def object_num(self):
        num = len(self.circle_list) + len(self.poly_list) 
        return num

    @property
    def 物體數量(self):
        num = len(self.circle_list) + len(self.poly_list) 
        return num


    @property
    def 模擬暫停(self):
        return self.pause_simulate

    @模擬暫停.setter
    def 模擬暫停(self, value):
        if value :
            self.pause_simulate = True
        else:
            self.pause_simulate = False 

    @property
    def 慢動作(self):
        return self.slow_simulate

    @慢動作.setter
    def 慢動作(self, value):
        if value:
            self.slow_simulate = True
        else:
            self.slow_simulate = False
  
    # @property
    # def 箭頭開始(self):
    #     return self.arrow_assist.enabled

    # @箭頭開始.setter
    # def 箭頭開始(self, value):
    #     if value:
    #         self.arrow_assist.enabled = True
    #     else:
    #         self.arrow_assist.enabled = False


import setuptools

# with open("README.md", "r",encoding="utf8") as fh:
#     long_description = fh.read()

setuptools.setup(
    name="pie4t",
    version="0.0.9",
    author="Wen-Hung, Chang 張文宏",
    author_email="beardad1975@nmes.tyc.edu.tw",
    description="Physics Impulse Engine wrapper for Teenagers",
    long_description="Physics Impulse Engine wrapper for Teenagers",
    long_description_content_type="text/markdown",
    url="https://github.com/beardad1975/pie4t",
    #packages=setuptools.find_packages(),
    platforms=["Windows"],
    python_requires=">=3.5",
    packages=['pie4t','物理模組'],
    install_requires = ['arcade~=2.6.0', 'pymunk~=6.2.0'],
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: Microsoft :: Windows",
            #"Operating System :: MacOS",
            #"Operating System :: POSIX :: Linux",
            "Natural Language :: Chinese (Traditional)",
        ],
)


import random
import math

import pymunk
from pymunk.vec2d import Vec2d
import arcade

from . import common


class PhysicsCommon:
    def limit_velocity(self, body, gravity, damping, dt):
        speeding = False
        l = body.velocity.length
        if l > common.VELOCITY_LIMIT:
            #print('found large velocity :', body.velocity)
            scale = common.VELOCITY_LIMIT / l
            body.velocity = body.velocity * scale
            speeding = True
            #print('fix to :', body.velocity)

        if not speeding :
            pymunk.Body.update_velocity(body, gravity, damping, dt)
        else:# speeding , no gravity
            pymunk.Body.update_velocity(body, (0,0), damping, dt)

    def limit_position(self, body, dt):
        # remove below lower bound
        pymunk.Body.update_position(body, dt)
        #print(body.position)
        x = body.position.x
        y = body.position.y
        if y < common.stage.win_height * common.RECYCLE_Y_MIN_MUL :
            #print('exceed y min')
            #body.position = Vec2d(x, common.POSITION_Y_MAX)
            common.stage.移除(self)
            #print('remove yVV: ')
        elif y > common.stage.win_height * common.RECYCLE_Y_MAX_MUL:
            #body.position = Vec2d(x, common.POSITION_Y_MIN )
            common.stage.移除(self)
            #print('remove y^^: ')
        elif x < common.stage.win_width * common.RECYCLE_X_MIN_MUL :
            
            #body.position = Vec2d(common.POSITION_X_MAX, y )
            common.stage.移除(self)
            #print('remove  <<x : ')
        elif x > common.stage.win_width * common.RECYCLE_X_MAX_MUL:
            #body.position = Vec2d(common.POSITION_X_MIN, y ) 
            common.stage.移除(self)
            #print('remove  x>> : ')       

            #print('exceed y max')
            #print(body.position)
            #common.stage.移除(self)
            #print('remove: ', self)

    def 施加力量(self, force):
        #self.phy_body.apply_force_at_world_point(force, (0,0))
        self.phy_body.force += force

    def 施加衝力(self, impulse):
        #self.phy_body.apply_impulse_at_world_point(impulse, (0,0))
        self.phy_body.velocity += impulse

    ### rigid body property
    @property
    def 質量(self):
        return self.phy_shape.mass

    @質量.setter
    def 質量(self, value):
        if value <= 0:
            print("質量須大於0")
        else:
            self.phy_shape.mass = value 

    @property
    def 密度(self):
        return self.phy_shape.density

    @密度.setter
    def 密度(self, value):
        if value <= 0:
            print("密度須大於0")
        else:
            self.phy_shape.density = value   


    @property
    def 位置(self):
        return self.phy_body.position

    @位置.setter
    def 位置(self, value):
        self.phy_body.position = value

    @property
    def 速度(self):
        return self.phy_body.velocity

    @速度.setter
    def 速度(self, value):
        self.phy_body.velocity = value

    @property
    def 角度(self):
        return math.degrees(self.phy_body.angle)

    @角度.setter
    def 角度(self, value):
        self.phy_body.angle = math.radians(value)

    @property
    def 角速度(self):
        return math.degrees(self.phy_body.angular_velocity)

    @角速度.setter
    def 角速度(self, value):
        self.phy_body.angular_velocity = math.radians(value)


    @property
    def 摩擦(self):
        return self.phy_shape.friction

    @摩擦.setter
    def 摩擦(self, value):
        if value < 0:
            print("摩擦係數須大於等於0")
        else:
            self.phy_shape.friction = value 

    @property
    def 彈性(self):
        return self.phy_shape.elasticity

    @彈性.setter
    def 彈性(self, value):
        if value < 0 or value > 1:
            print("彈性係數須在0到1之間")
        else:
            self.phy_shape.elasticity = value    

    @property
    def 物理反應(self):
        if self.phy_body.body_type == pymunk.Body.DYNAMIC:
            return True
        elif self.phy_body.body_type == pymunk.Body.KINEMATIC:
            return False

    @物理反應.setter
    def 物理反應(self, g):
        if g is True and self.phy_body.body_type ==  pymunk.Body.KINEMATIC:
            # change type to dynamic
            self.phy_body.body_type = pymunk.Body.DYNAMIC
            if common.stage.is_engine_running:
                self.current_shape_element = self.dynamic_shape_element
        elif  g is False and self.phy_body.body_type ==  pymunk.Body.DYNAMIC:
            # chang type to kinematic
            self.phy_body.body_type = pymunk.Body.KINEMATIC
            if common.stage.is_engine_running:
                self.current_shape_element = self.kinematic_shape_element






import pymunk
import arcade

from . import common


class Segment:
    def __init__(self, a, b, 寬=None):
        self.is_lazy_setup = False



        # pymunk part
        self.a = a
        self.b = b

        if 寬 is not None:
            self.thickness = 寬
        else:
            self.thickness = common.SEG_THICKNESS

        self.phy_body = pymunk.Body(body_type=pymunk.Body.STATIC)
        self.phy_body.position = a
        a_b_delta = (b[0] - a[0], b[1]-a[1])
        self.phy_shape = pymunk.Segment(self.phy_body,(0,0), a_b_delta, 
                                self.thickness//2)
        self.phy_shape.friction = 1
        self.phy_shape.elasticity = 1
        self.phy_shape.collision_type = common.COLLITYPE_SEGMENT

        self.phy_shape.filter = pymunk.ShapeFilter(categories=common.CATE_SEGMENT)

        self.phy_shape.obj = self

        common.stage.space.add(self.phy_body)
        common.stage.space.add(self.phy_shape)
        
        if common.stage.is_engine_running:
            self.lazy_setup()
            common.stage.is_engine_running = True

    def lazy_setup(self):
        if not self.is_lazy_setup:
            #arcade part
            self.shape_element = arcade.ShapeElementList()
            a = self.a
            b = self.b
            self.line_shape = arcade.create_line(a[0], a[1], b[0], b[1],
                     arcade.color.YELLOW_ORANGE, self.thickness)
            self.is_lazy_setup = True

    def draw(self):
        self.line_shape.draw()
        
    